/* ------------------------------- calcsigma.c ----------------------- *
 
 Written by BWO, 2006 (updated 25 January 2007).

 This routine takes in cosmological parameters and creates a file with
 sigma(M) in it, which is necessary for various press-schechter type
 stuff.  In principle one can calculate it ahead of time, but it's far,
 far faster in the long run to calculate your sigma(M) ahead of time.

 Inputs: cosmology, possibly a transfer function if the user is doing
   CMBFAST.  User must set parameters below.  CMBFAST transfer function
   is generated separately, and the user has to figure out how to do that.

 Outputs: lots of stdout, plus four columns of data with a filename
   specified by the user, containing the following information:

   1) log mass (Msolar)
   2) mass (Msolar/h)
   3) Radius (comoving Mpc/h)
   4) sigma (normalized) using Msun/h as the input

 * ------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_integration.h>
#include <time.h>

/* ----------------- USER SETS ALL OF THESE THINGS ------------------- */

char *outfile = "WMAP5_s8_0_817.dat";

// WMAP YEAR 5 recommended (WMAP+BAO+SNe)
double omega_matter0 = 0.279,
  omega_lambda0 = 0.721,
  omega_baryon0 = 0.0462,
  hubble0 = 0.701,
  sigma8input = 0.817,
  primordial_index = 0.960;


// our "fiducial" cosmology runs
/*
double omega_matter0 = 0.3,
  omega_lambda0 = 0.7,
  omega_baryon0 = 0.04, 
  hubble0 = 0.7,
  sigma8input = 0.9,
  primordial_index = 1.0;
*/

// WMAP YEAR III best-fit
/*
double omega_matter0 = 0.268,
  omega_lambda0 = 0.732,
  omega_baryon0 = 0.0441,
  hubble0 = 0.704,
  sigma8input = 0.85,
  primordial_index = 0.947;
*/



// not cosmology, but must be set anyway.
double
  log_mass_min = 2.0,  // in units of Msolar, NOT Msolar/h
  log_mass_max = 20.0, // in units of Msolar, NOT Msolar/h
  num_sigma_bins = 360.0;  // make sure this has a decimal point

int transfer_function_type = 2;  // 1 = bardeen
                                 // 2 = Eisenstein & Hu
                                 // 3 = transfer function from CMBFAST 
                                 //     (assumes standard 5-column file format -
                                 //      see below for more parameters to set)

double allowed_error = 1.0e-6;  // allowed fractional error in numerical integration.
                                // recommended vals of 1e-4 for CMBFAST, 1e-6 for others!

// CMBFAST REQUIRED PARAMETERS (these only need to be set if 
// transfer_function_type is 3)
char *cmbfastfilename = "cmb_transfer_fctn";  // file name of actual CMBFAST file

double cmbfast_min_k = 0.188374E-04, // first row, first column in file
  cmbfast_max_k =  0.111627E+05,     // last row, first column in file
  cmbfast_total_bins = 102;          // total lines in file

/* ---------------------- END USER SET ---------------------- */

double sigma_normalization;

double *cmbfast_kvals, *cmbfast_TofKvals;

// set to read in the CMBFAST file (with parameters defined above)
int read_cmbfast_file(void);


double sigma_squared_of_R(double R);

// various routines for Eisenstein & Hu transfer function
int TFmdm_set_cosm(double omega_matter, double omega_baryon, double omega_hdm,
	int degen_hdm, double omega_lambda, double hubble, double redshift);
double TFmdm_onek_mpc(double kk);
double TFmdm_onek_hmpc(double kk);


int main (void)
{
  double R, sigma8_unnorm, thismass, thislogmass, thisradius, rho0, dm;
  double *Rarray, *logmassarray, *massarray, *sigmaarray;
  int i, returnnumlines;
  time_t  start_time, end_time, total_time;

  FILE *fptr;

  start_time = time(NULL);

  int errcode;

  // initiate E&H transfer function if that is what the user requests
  if(transfer_function_type == 2)
    errcode = TFmdm_set_cosm(omega_matter0, omega_baryon0, 0.0, 0, 
			     omega_lambda0, hubble0, 0.0);

  // read in CMBFAST transfer function if that is what the user requests,
  // into arrays declared for the purpose.
  // note: CMBFAST k values are in units of k/h
  if(transfer_function_type == 3){
    cmbfast_kvals = new double[ (int) cmbfast_total_bins ];
    cmbfast_TofKvals = new double[ (int) cmbfast_total_bins ];;

    returnnumlines = read_cmbfast_file();
  }

  // output arrays
  Rarray = new double[ (int) num_sigma_bins ];
  logmassarray = new double[ (int) num_sigma_bins ];
  massarray = new double[ (int) num_sigma_bins ];
  sigmaarray = new double[ (int) num_sigma_bins ];

  // get sigma_8 normalization

  R = 8.0;  // in units of Mpc/h (comoving)

  sigma8_unnorm = sqrt(sigma_squared_of_R(R));

  sigma_normalization = sigma8input / sigma8_unnorm;

  printf("sigma8 unnormalized:  %e\n",sigma8_unnorm);
  printf("sigma8 input:  %e\n",sigma8input);
  printf("normalization:  %e\n",sigma_normalization);

  rho0 = omega_matter0 * 2.78e+11; // in units of h^2 Msolar/Mpc^3

  // spacing in mass of our sigma calculation
  dm = (log_mass_max - log_mass_min)/num_sigma_bins;

  fptr = fopen(outfile, "w");

  /* loop over the total number of sigma_bins the user has requested. 
     For each bin, calculate mass and equivalent radius, and call
     sigma_squared_of_R to get the sigma(R) (equivalent to sigma(M)),
     normalize by user-specified sigma_8, and then write out. */
  for( i = 0; i < ((int) num_sigma_bins ) ; i++){

    // thislogmass is in units of Msolar, NOT Msolar/h
    thislogmass = log_mass_min + ( (double) i )*dm; // + double(i)*dm;

    // mass in units of h^-1 Msolar
    thismass = pow(10.0, thislogmass) * hubble0; 

    // radius is in units of h^-1 Mpc (comoving)
    thisradius = pow( 3.0*thismass / 4.0 / M_PI / rho0, 1.0/3.0 );

    R = thisradius; // h^-1 Mpc (comoving)

    Rarray[i] = thisradius;  // h^-1 Mpc (comoving)
    logmassarray[i] = thislogmass;  // Msun (NOT Msun/h)
    massarray[i] = thismass;  // Msun/h

    // get normalized sigma(R)
    sigmaarray[i] = sqrt(sigma_squared_of_R(R)) * sigma_normalization;

    // print out values:
    // 1) log mass (Msolar)
    // 2) mass (Msolar/h)
    // 3) Radius (comoving Mpc/h)
    // 4) sigma (normalized, using Msun/h as the input)

    fprintf(fptr, "%lf %e %lf %lf\n", logmassarray[i], 
	    massarray[i], Rarray[i], sigmaarray[i]);

    
    printf("%lf %e %lf %lf\n", logmassarray[i], 
	    massarray[i], Rarray[i], sigmaarray[i]);
    
  } // for( i = 0; i < ((int) num_sigma_bins ) ; i++)

  fclose(fptr);

  delete [] Rarray;
  delete [] logmassarray;
  delete [] sigmaarray;
  delete [] massarray;

  if(transfer_function_type == 3){
    delete [] cmbfast_kvals;
    delete [] cmbfast_TofKvals;
  }

  end_time = time(NULL);

  total_time = (time_t) difftime(end_time,start_time);

  printf("this required %lf  seconds to run (%lf per sigma calculation).\n",
	 (double) total_time,  ((double) total_time) / num_sigma_bins);

  return 0;
}


/* simple routine to read in the CMBFAST data file.  Assumes standard CMBFAST transfer
   function output style (no user mods necessary).  Returns the number of lines read. */
int read_cmbfast_file(void){

  FILE *fptr;
  int lines=0;
  double djunk;

  fptr = fopen(cmbfastfilename,"r");

  while(fscanf(fptr,"%lf %lf %lf %lf %lf",&cmbfast_kvals[lines],
	       &cmbfast_TofKvals[lines],&djunk,&djunk,&djunk)==5)
    lines++;

  fclose(fptr);

  printf("read in file %s, total of %d lines read in.\n",cmbfastfilename,lines);
  printf("    first line is:   %e  %e\n",cmbfast_kvals[0],cmbfast_TofKvals[0]);
  printf("    last line is:    %e  %e\n",cmbfast_kvals[lines-1], cmbfast_TofKvals[lines-1]);
  fflush(stdout);

  return lines;
}


/* calculates sigma^2(R).  This is the routine where the magic happens (or
   whatever it is that we do here).  Integrates the sigma_squared_integrand
   parameter from R to infinity.  Calls GSL (gnu scientific library) to do
   the actual integration.  

   Note that R is in h^-1 Mpc (comoving)
*/
double sigma_squared_of_R(double R){

  double sigma_squared_integrand(double k, void * params);

  gsl_integration_workspace * w 
    = gsl_integration_workspace_alloc (10000);
  
  double result, error, sigmasquaredofR;

  gsl_function F;

  F.function = &sigma_squared_integrand;
  F.params = &R;

  gsl_integration_qagiu (&F, 0.0, allowed_error, allowed_error, 10000,
			w, &result, &error); 

  /*  
  printf ("result          = % .18f\n", result);
  printf ("estimated error = % .18f\n", error);
  printf ("intervals =  %d\n", (int) w->size);
  printf ("R               = %.18lf\n",R);
  */

  sigmasquaredofR = result / 2.0 / M_PI / M_PI;

  return sigmasquaredofR;
}

/* integrand for integral to get sigma^2(R). */
double sigma_squared_integrand(double k, void * params) {

  double PofK(double k);
  double WofK(double R, double k);
  double f;

  double Rcom = *(double *) params;  // this is R in comoving Mpc/h

  f = k*k*PofK(k)*pow( fabs(WofK(Rcom,k)), 2.0);

  return f;
}

/* returns power spectrum as a function of wavenumber k */
double PofK(double k){

  double TofK(double k);

  double thisPofK;

  thisPofK = pow(k, primordial_index) * pow( TofK(k), 2.0);

  return thisPofK;

}

/* returns transfer function as a function of wavenumber k.  This is
   sort of an ugly function since there are so many user-specifiable 
   transfer functions. */
double TofK(double k){

  double TofK_cmbfast(double k);

  double q, gamma, thisTofK;
    
  if(transfer_function_type == 1){ // Bardeen fitting function (ancient and basically wrong)
 
    gamma = omega_matter0 * hubble0 * 
      exp( -1.0*omega_baryon0 - omega_baryon0 / omega_matter0);
    
    q = k / (hubble0*gamma);

    thisTofK = log(1.0 + 2.34*q) / 2.34 * 
      pow( 1.0 + 3.89*q + pow(16.1*q, 2.0) + pow(5.46*q, 3.0) + pow(6.71*q, 4.0)
	   , -0.25);

  } else if(transfer_function_type == 2) {  // Eisenstein & Hu fitting function
    
    thisTofK = TFmdm_onek_hmpc(k);
    
  } else if(transfer_function_type == 3) {  // T(k) right from CMBFAST

    thisTofK = TofK_cmbfast(k);

  } else {  // error!

    printf("I don't know about transfer function type %d.  Exiting.",transfer_function_type);
    exit(-111);

  }

  return thisTofK;

}


// returns W(k), which is the fourier transform of the top-hat function.
double WofK(double R, double k){

  double thisWofK, x;

  x = R*k;

  thisWofK = 3.0 * ( sin(x) - x*cos(x) ) / (x*x*x);

  //printf("thisWofK:  %e\n", thisWofK);

  return thisWofK;

}

// calculates T(k) for CMBFAST.  This could probably
// be sped up a bit.
double TofK_cmbfast(double k){

  double cmbfast_tofk;

  double alpha=0.0;

  cmbfast_tofk = -1.0;


  // Try to set T(k) using simple linear interpolation.  This
  // could probably be improved by using a spline interpolation
  // but I don't have a spline routine laying around and I'm
  // busy.
  for(int i=0; i < ((int) cmbfast_total_bins)-1; i++)
    if( (cmbfast_kvals[i] < k) && (k <= cmbfast_kvals[i+1]) )
      cmbfast_tofk = cmbfast_TofKvals[i] 
	+ (k-cmbfast_kvals[i]) / ( cmbfast_kvals[i+1]-cmbfast_kvals[i] )    
	* ( cmbfast_TofKvals[i+1] - cmbfast_TofKvals[i] ) ;

  // if the values are below or above the transfer functions given by cmbfast,
  // we have to make some estimates of T(k).
  if( cmbfast_tofk < 0.0){

    // transfer function is flat at low k (large spatial scales)
    if( k <= cmbfast_kvals[0]) cmbfast_tofk = cmbfast_TofKvals[0];

    // Transfer function is roughly a power law at high k, so we
    // make a crude extrapolation.
    if( k > cmbfast_kvals[ ((int) cmbfast_total_bins)-2]){

      // take logarithmic derivative of trasnfer fctn to get power of power law, alpha

      alpha = log10(  cmbfast_TofKvals[((int) cmbfast_total_bins)-1] 
		      /  cmbfast_TofKvals[((int) cmbfast_total_bins)-4] )
	/ log10(  cmbfast_kvals[((int) cmbfast_total_bins)-1] 
		  /  cmbfast_kvals[((int) cmbfast_total_bins)-4] );


      // calculate T(k) using that power law and the last value of the CMBFAST
      // table:  T(k) = T(k_end)*(k/k_end)^alpha
      cmbfast_tofk = cmbfast_TofKvals[((int) cmbfast_total_bins)-1] 
	* pow( k / cmbfast_kvals[((int) cmbfast_total_bins)-1] , alpha );
	
    }

    // error check.
    if(cmbfast_tofk < 0.0){
      fprintf(stderr,"dammit, this should never be reached. 3321.\n");
      exit(-121);
    }

  } // if( cmbfast_tofk < 0.0)

  return cmbfast_tofk;

}

/* done! */
