/*  Press-Schecter mass function code.

  outputs are (redshift, number density) where number density is per Mpc^3 (NOT h^3 / Mpc^3)

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

/* ------------------ USER SETS ALL OF THESE ------------------ */

char *sigmafile = "fiducial_sigma8_0_9_om_0_3_ol_0_7.dat";
char *dndmfile = "warren_dncum_z_fiducial_sigma8_0_9_om_0_3_ol_0_7.dat";

double 

//  our "fiducial" cosmology 

  omega_lambda0 = 0.7,
  omega_matter0 = 0.3,
  omega_baryon0 = 0.04,
  hubble0 = 0.7,
  primordial_index = 1.0,



/*
// WMAP YEAR III 
  omega_lambda0 = 0.732,
  omega_matter0 = 0.268, 
  omega_baryon0 = 0.0441, 
  hubble0 = 0.704, 
  primordial_index = 0.947,
*/

  log_mass_min = 2.0,  // in units of Msolar, NOT Msolar/h
  log_mass_max = 20.0, // in units of Msolar, NOT Msolar/h
  num_sigma_bins = 360.0,
  z_min = 0.0,
  z_max = 5.0,
  z_bins = 360.0,
  threshold_mass = 1.4286e+14  // in Msolar, NOT Msolar/h
  ; // make sure this has a decimal point

int fitting_function = 4;  // 1 = Press-schechter
                           // 2 = Jenkins
                           // 3 = Sheth-Tormen
                           // 4 = LANL (Warren) fit



/* ------------------------------------------------------------ */

double sigmaof_M_z(int sigmabin, double redshift);
double multiplicityfunction(double sigma);
 
double rho0, *Rarray, *logmassarray, *massarray, *sigmaarray, 
  nu, delta_c0, delta_c_of_z, thisredshift;

int main (void)
{
  int i;
  double nofmz_cum, dn_M_z, thissigma, nextsigma, dsigmadm;
  FILE *fptr;

  Rarray = new double[ (int) num_sigma_bins ];
  logmassarray = new double[ (int) num_sigma_bins ];
  massarray = new double[ (int) num_sigma_bins ];
  sigmaarray = new double[ (int) num_sigma_bins ];


 // read in values from sigma(M) file.  This must be in the following format:
  // 1) log mass (Msolar, NOT Msolar/h)
  // 2) mass (Msolar/h)
  // 3) Radius (comoving Mpc/h)
  // 4) sigma (normalized and using Msun/h as the input)

  i=0;

  fptr = fopen(sigmafile,"r");

  while(fscanf(fptr, "%lf %lf %lf %lf", &logmassarray[i], 
	       &massarray[i], &Rarray[i], &sigmaarray[i])==4) i++;

  printf("read in %d lines from file %s\n",i,sigmafile);

  fclose(fptr);

  // constants - set these before calling any functions!
  rho0 = omega_matter0 * 2.78e+11; // in units of h^2 Msolar/Mpc^3
  delta_c0 = 1.69;  // critical density for turnaround (Press-Schechter)

  double dz;

  dz = (z_max - z_min) / z_bins;

  // file pointer to output file
  fptr = fopen(dndmfile,"w");



  for(double thisz=z_min; thisz <= z_max; thisz += dz){

    thisredshift = thisz;

    nofmz_cum = 0.0;  // keep track of cumulative number density

    // Loop over masses, going BACKWARD, and calculate dn/dm as well as the 
    // cumulative mass function.

    for(i = ((int) num_sigma_bins)-2; i >= 0; i--){

      thissigma = sigmaof_M_z(i,thisredshift);

      nextsigma = sigmaof_M_z(i+1,thisredshift);

      // calc dsigmadm - has units of h (since massarray has units of h^-1)
      dsigmadm = (nextsigma-thissigma) / (massarray[i+1] - massarray[i]);

      // calculate dn(M,z) (dn/dM * dM)
      // this has units of h^3 since rho0 has units of h^2, dsigmadm has units of h, and massarray has units of h^-1
      dn_M_z = -1.0 / thissigma * dsigmadm * rho0 / massarray[i] * multiplicityfunction(thissigma)*(massarray[i+1] - massarray[i]);

      // scale by h^4 to get rid of all factors of h
      dn_M_z *= pow(hubble0, 4.0);

      // keep track of cumulative number density
      if( (massarray[i] > threshold_mass*hubble0) &&  (dn_M_z > 1.0e-20) )
	nofmz_cum += dn_M_z;

    } //   for(i = ((int) num_sigma_bins)-2; i >= 0; i--){

    fprintf(fptr,"%lf %e\n",thisredshift,nofmz_cum);

  } //   for(double thisz=z_min; thisz <= zmax; this += dz)

  fclose(fptr);

  delete [] Rarray;
  delete [] logmassarray;
  delete [] sigmaarray;
  delete [] massarray;

  return 0;
}


/* Multiplicity function - this is where the various fitting functions/analytic 
   theories are different.  The various places where I found these fitting functions
   are listed below.  */
double multiplicityfunction(double sigma){

  double thismult, nu, A, a, b, c, p;

  nu = delta_c0 / sigma;

  switch(fitting_function){

  case 1:  // Press-Schechter (This form from Jenkins et al. 2001, MNRAS 321, 372-384, eqtn. 5)
    thismult = sqrt(2.0/M_PI) * nu * exp(-0.5*nu*nu);
    break;

  case 2:  // Jenkins et al. 2001, MNRAS 321, 372-384, eqtn. 9
    thismult = 0.315 * exp( -1.0 * pow( fabs( log(1.0/sigma) + 0.61), 3.8 ) );
    break;

  case 3:  // Sheth-Tormen 1999, eqtn 10, using expression from Jenkins et al. 2001, eqtn. 7
    A=0.3222;
    a=0.707;
    p=0.3;
    thismult = A*sqrt(2.0*a/M_PI)*(1.0+ pow( 1.0/(nu*nu*a), p) ) * nu * exp(-0.5*a*nu*nu);
    break;

  case 4:  // LANL fitting function - Warren et al. 2005, astro-ph/0506395, eqtn. 5 
    A=0.7234; 
    a=1.625; 
    b=0.2538; 
    c=1.1982;
    thismult = A*( pow(sigma, -1.0*a) + b)*exp(-1.0*c / sigma / sigma );
    break;

  default:  // if we don't have one in here
    fprintf(stderr,"Don't understand this.  FItting function requested is %d\n",fitting_function);
    exit(-1);
    break;
   
  }  // switch(fitting_function){

  return thismult;

}


/* sigma(M, z) */
double sigmaof_M_z(int sigmabin, double redshift){

  double thissigma;
  
  double Dofz(double redshift);

  thissigma = Dofz(redshift) * sigmaarray[sigmabin];

  return thissigma;

}


/* Growth function */
double Dofz(double redshift){

  double gofz(double redshift);
  double thisDofz;

  thisDofz = gofz(redshift) / gofz(0.0) / (1.0+redshift);

  return thisDofz;
}

/* g(z) - I don't think this has any other name*/
double gofz(double redshift){

  double thisgofz;
  double omega_matter_of_z(double redshift);
  double omega_lambda_of_z(double redshift);

  thisgofz = 2.5 * omega_matter_of_z(redshift) / 
    ( pow( omega_matter_of_z(redshift), 4.0/7.0 )
      - omega_lambda_of_z(redshift)
      + ( (1.0 + omega_matter_of_z(redshift) / 2.0)
	  * (1.0 + omega_lambda_of_z(redshift) / 70.0) )
      );

  return thisgofz;
}

/* Omega matter as a function of redshift */
double omega_matter_of_z(double redshift){
  double thisomofz;
  double Eofz(double redshift);

  thisomofz = omega_matter0 * pow( 1.0+redshift, 3.0) 
    / pow( Eofz(redshift), 2.0 );

  return thisomofz;
}

/* Omega lambda as a function of redshift */
double omega_lambda_of_z(double redshift){
  double thisolofz;
  double Eofz(double redshift);

  thisolofz = omega_lambda0 / pow( Eofz(redshift), 2.0 );

  return thisolofz;
}

/* E(z) - I don't think this has any other name */
double Eofz(double redshift){
  double thiseofz;

  thiseofz = sqrt( omega_lambda0
		  + (1.0 - omega_lambda0 - omega_matter0)*pow( 1.0+redshift, 2.0)
		   + omega_matter0 * pow( 1.0+redshift, 3.0)  );

  return thiseofz;
}
