/* ------------------------------- calcsigma.c ----------------------- *
 
 Written by BWO, Feb 2007 

  Takes in cosmological parameters and an angular opening and creates a 
  file with the cosmological comoving volume element as a function of 
  redshift.

   Inputs: cosmology, redshift values, dz.  User must set parameters below.  

 Outputs: lots of stdout, plus two columns of data with a filename
   specified by the user, containing the following information:

   1) redshift
   2) cosmological comoving volume element, in (Mpc/h)^3

 * ------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_integration.h>
#include <time.h>

/* ----------------- USER SETS ALL OF THESE THINGS ------------------- */

char *outfile = "volume_fiducial_sigma8_0_9_om_0_4_ol_0_6.dat";

// our "fiducial" cosmology runs
double omega_matter0 = 0.4,
  omega_lambda0 = 0.6,
  omega_baryon0 = 0.04,
  hubble0 = 0.7,
  primordial_index = 1.0;

/*
// WMAP YEAR III best-fit
double omega_matter0 = 0.268,
  omega_lambda0 = 0.732,
  omega_baryon0 = 0.0441,
  hubble0 = 0.704,
  primordial_index = 0.947;
*/


// not cosmology, but must be set anyway.
double
  z_min = 0.0,  // in units of Msolar, NOT Msolar/h
  z_max = 5.0, // in units of Msolar, NOT Msolar/h
  num_z_bins = 360.0;  // make sure this has a decimal point


double allowed_error = 1.0e-4;  // allowed fractional error in numerical integration.

/* ---------------------- END USER SET ---------------------- */

int main (void)
{
  double *redshiftarray, *volumearray,
    thisredshift, dz;

  FILE *fptr;

  double volume(double redshift);

  // output arrays
  redshiftarray = new double[ (int) num_z_bins ];
  volumearray = new double[ (int) num_z_bins ];

  // spacing in redshift of our volume calculation
  dz = (z_max - z_min)/num_z_bins;

  fptr = fopen(outfile, "w");

  /* loop over the total number of sigma_bins the user has requested. 
     For each bin, calculate mass and equivalent radius, and call
     sigma_squared_of_R to get the sigma(R) (equivalent to sigma(M)),
     normalize by user-specified sigma_8, and then write out. */
  for(int i = 0; i < ((int) num_z_bins ) ; i++){

    thisredshift = z_min + ( (double) i )*dz; // + double(i)*dm;

    redshiftarray[i] = thisredshift;

    // get volume array (per square degree)
    volumearray[i] = volume(thisredshift) / 3282.806;

    // print out values:
    // 1) redshift
    // 2) volume

    fprintf(fptr, "%lf %e\n", redshiftarray[i], 
	    volumearray[i]);
    
  } // for( i = 0; i < ((int) num_z_bins ) ; i++)

  fclose(fptr);

  delete [] redshiftarray;
  delete [] volumearray;

  return 0;
}


double volume(double redshift){
  double Eofz(double redshift);
  double volume_integrand(double redshift, void * params);

  gsl_integration_workspace * w 
    = gsl_integration_workspace_alloc (10000);
  
  double result, error, thisvol, angdiadist;

  gsl_function F;

  F.function = &volume_integrand;

  gsl_integration_qag (&F, 0.0, redshift, allowed_error, allowed_error, 10000,
		       GSL_INTEG_GAUSS61,
			w, &result, &error); 

  angdiadist = result * 3000.0 / hubble0 / (1.0+redshift);

  //thisvol = 3000.0 / hubble0 / Eofz(redshift) * pow(angdiadist, 2.0) * pow( (1.0+redshift), 2.0);

  thisvol = pow( (3000.0/hubble0), 3.0) * pow(result, 2.0) / Eofz(redshift);


  printf("%lf  %e   %e\n",redshift, Eofz(redshift), thisvol);


  return thisvol;
}

double volume_integrand(double redshift, void * params) {

  double Eofz(double redshift);
  double f;

  f = 1.0 / Eofz(redshift);

  return f;
}

/* E(z) - I don't think this has any other name */
double Eofz(double redshift){
  double thiseofz;

  thiseofz = sqrt( omega_lambda0
		  + (1.0 - omega_lambda0 - omega_matter0)*pow( 1.0+redshift, 2.0)
		   + omega_matter0 * pow( 1.0+redshift, 3.0)  );

  return thiseofz;
}
